% program to find  the frequency response of a Doherty amplifier
% with predetermined output backoff levels
% Enter the parameters in DohertyInit.m and run that program first to
% generate BOn.mat, which will be used as input to this program
% Using the notation of Alemdar & Atalar paper
% Using the new odd+even model
clear


tic
load('BOn.mat')
wN=fN*2*pi*1e9;

% --------------------------------------------
% Select Start Stop Frequency
wmax=1.2*wN; %Stop Frequency
wmin=0.8*wN; %Start Frequency

nw=61; % number of frequency steps of input excitation
% end of input parameters
% ------------------------------------------

w=wmin:(wmax-wmin)/(nw-1):wmax;
fn=w/(2*pi*1e9); % the actual frequencies in GHz

% initialize matrices to reserve space
POUT=zeros(1,nw); 
GAIN=POUT;VDSm=POUT;VDSp=POUT;IDSm=POUT;IDSp=POUT;
I0m=POUT;I0p=POUT;Zm=POUT;Zp=POUT;ZL=POUT;
PDB=zeros(NoOfSteps,nw);
EFF=POUT;
Am=zeros(2,2,nw);
Ap=Am;
AOMN=Ap;
ZL1=zeros(1,1,nw);

myfun=@sA1;
% options for nonlinear solver
options = optimoptions('fsolve','Display','none','OptimalityTolerance',1e-7);
ig=1;
IDQm=Gm*(VBm-VT);
IDQp=Gm*(VBp-VT);
figure(4)
subplot(1,2,1)
hold off
subplot(1,2,2)
hold off
TotPow=0;
TotEff=0;

stepNo=0;
for ii=1:NoOfSteps
    Vin=Ipmax/Gm*10^(BOn(ii)/20);
    Vinm=Vin*alpha;
    Vinp=Vin*beta;

    Ipm=Gm*(Vinm+VBm-VT);
    Ipp=Gm*(Vinp+VBp-VT);

    mFVAL=0;
    % find the ABCD matrices
    % ABCDm for main 
    % ABCDp for peaking
    for iw=1:nw
    switch ConvDo
    case 1  %if conventional Doherty amplifier (Conv)
            Ap(:,:,iw)=[1 0; 1i*w(iw)*C 1]*[1 1i*w(iw)*Ls; 0 1]*[1 0;1/(1i*w(iw)*Lsh) 1];
            Am(:,:,iw)=[1 0; 1i*w(iw)*C 1]*[1 1i*w(iw)*Ls; 0 1]*[1 0;1/(1i*w(iw)*Lsh) 1]*[cos(pi/2*w(iw)/wN) 1i*Z2*sin(pi/2*w(iw)/wN);1i/Z2*sin(pi/2*w(iw)/wN) cos(pi/2*w(iw)/wN)]; 
    case 0  %if two section TL (type I in paper)
            Am(:,:,iw)=[1 0; 1i*w(iw)*C 1]*[1 1i*w(iw)*Ls; 0 1]*[1 0;1/(1i*w(iw)*Lsh) 1]*[cos(pi/2*w(iw)/wN) 1i*Z2*sin(pi/2*w(iw)/wN);1i/Z2*sin(pi/2*w(iw)/wN) cos(pi/2*w(iw)/wN)];  
            Ap(:,:,iw)=[1 0; 1i*w(iw)*C 1]*[1 1i*w(iw)*Ls; 0 1]*[1 0;1/(1i*w(iw)*Lsh) 1]*[cos(pi/2*w(iw)/wN) 1i*Z3*sin(pi/2*w(iw)/wN);1i/Z3*sin(pi/2*w(iw)/wN) cos(pi/2*w(iw)/wN)]*[cos(pi/2*w(iw)/wN) 1i*Z4*sin(pi/2*w(iw)/wN);1i/Z4*sin(pi/2*w(iw)/wN) cos(pi/2*w(iw)/wN)];
    case 8  %if C-L-C inverter (type II in paper)
            Ap(:,:,iw)=[1 0; 1i*w(iw)*C 1]*[1 1i*w(iw)*Ls; 0 1]*[1 0;1/(1i*w(iw)*Lsh) 1]*[1 0; 1i*w(iw)*C4 1]*[1 0; -1i/(w(iw)*L3) 1]*[1 1i*w(iw)*L2; 0 1]*[1 0; 1i*w(iw)*C3 1]*[1 0; 1i*w(iw)*C_offset 1]*[1 1i*w(iw)*L_offset; 0 1]*[1 0; 1i*w(iw)*C_offset 1];
            Am(:,:,iw)=[1 0; 1i*w(iw)*C 1]*[1 1i*w(iw)*Ls; 0 1]*[1 0;1/(1i*w(iw)*Lsh) 1]*[1 0; 1i*w(iw)*C2 1]*[1 0; -1i/(w(iw)*L1) 1]*[1 1i*w(iw)*L; 0 1]*[1 0; 1i*w(iw)*C1 1];
    case {5,6} %if C-TL-C inverter (type III in paper)
            Ap(:,:,iw)=[1 0; 1i*w(iw)*C 1]*[1 1i*w(iw)*Ls; 0 1]*[1 0;1/(1i*w(iw)*Lsh) 1]*[1 0; 1i*w(iw)*C4 1]*[1 0; -1i/(w(iw)*L3) 1]*[cos(x1*2*pi*w(iw)/wN) 1i*Z3*sin(x1*2*pi*w(iw)/wN); 1i/Z3*sin(x1*2*pi*w(iw)/wN) cos(x1*2*pi*w(iw)/wN)]*[1 0; 1i*w(iw)*C3 1]*[cos(2*y*pi*w(iw)/wN) 1i*Z4*sin(2*y*pi*w(iw)/wN);1i/Z4*sin(2*y*pi*w(iw)/wN) cos(2*y*pi*w(iw)/wN)];
            Am(:,:,iw)=[1 0; 1i*w(iw)*C 1]*[1 1i*w(iw)*Ls; 0 1]*[1 0;1/(1i*w(iw)*Lsh) 1]*[1 0; 1i*w(iw)*C2 1]*[1 0; -1i/(w(iw)*L1) 1]*[cos(x*2*pi*w(iw)/wN) 1i*Z2*sin(x*2*pi*w(iw)/wN); 1i/Z2*sin(x*2*pi*w(iw)/wN) cos(x*2*pi*w(iw)/wN)]*[1 0; 1i*w(iw)*C1 1];
    end           

        
    

    Zout=ZOUT; % corresponding normalized load impedance of Doherty
 
    AOMN(:,:,iw)=...
        ...%shunt component 1
        [1 0; 1i*w(iw)*CA 1]*...
        ...%series component 1
        [cos(y1*2*pi*w(iw)/wN) 1i*Z5*sin(y1*2*pi*w(iw)/wN); 1i/Z5*sin(y1*2*pi*w(iw)/wN) cos(y1*2*pi*w(iw)/wN)]*...
        [1 1i*w(iw)*LA; 0 1]*...
        ...%shunt component 2
        [1 0; 1i*w(iw)*(CB+CC) 1]*...
        ...%series component 2
        [cos(y2*2*pi*w(iw)/wN) 1i*Z6*sin(y2*2*pi*w(iw)/wN); 1i/Z6*sin(y2*2*pi*w(iw)/wN) cos(y2*2*pi*w(iw)/wN)]*...
        [1 1i*w(iw)*LB; 0 1]*...
        ...%shunt component 3
        [1 0; 1i*w(iw)*CD 1];
    
    ZL1(1,1,iw)=(AOMN(1,1,iw)*ZOUT+AOMN(1,2,iw))/(AOMN(2,1,iw)*ZOUT+AOMN(2,2,iw));
    end
    ZL(1,:)=ZL1(1,1,:);


% find the coefficients to be used later for all ZL values
    Den=Am(2,1,:).*Ap(2,2,:)+Am(2,2,:).*Ap(2,1,:);
    Den(1,1,:)=Am(2,2,:).*Ap(2,2,:)+Den(1,1,:).*ZL1(1,1,:);
    A11=(Am(1,2,:).*Ap(2,2,:)+(Am(1,1,:).*Ap(2,2,:)+Am(1,2,:).*Ap(2,1,:)).*ZL1(1,1,:))./Den(1,1,:);
    A12=((Am(1,1,:).*Am(2,2,:)-Am(1,2,:).*Am(2,1,:)).*ZL1(1,1,:))./Den(1,1,:);
    A22=(Ap(1,2,:).*Am(2,2,:)+(Ap(1,1,:).*Am(2,2,:)+Ap(1,2,:).*Am(2,1,:)).*ZL1(1,1,:))./Den(1,1,:);
    A21=((Ap(1,1,:).*Ap(2,2,:)-Ap(1,2,:).*Ap(2,1,:)).*ZL1(1,1,:))./Den(1,1,:);
    B11=ZL1(1,1,:).*Ap(2,2,:)./Den(1,1,:);
    B12=ZL1(1,1,:).*Am(2,2,:)./Den(1,1,:);
    B11a(1,:)=B11(1,1,:);
    B12a(1,:)=B12(1,1,:);
    

    VDSi=[0.2177-1i*0.9535;0.6511-1i*0.1119];
    options=optimoptions('fsolve','Display','off','MaxIterations',9000,'Algorithm','levenberg-marquardt','StepTolerance',1e-9,'MaxFunctionEvaluations',90000,'FunctionTolerance',1e-7);

    for ip=1:nw
        AT=[A11(1,1,ip) A12(1,1,ip);A21(1,1,ip) A22(1,1,ip)];
        wp=w(ip)/wN;
        
        itercnt=1;
        COND=true;
        while COND
        fun=@(VDS) myfun(Ipm,Ipp,IDQm,IDQp,VDS,No,nchko,Ne,nchke,AT,wp,VDD,ConvDoh,y);
        [VDS,FVAL,EXITFLAG]=fsolve(fun,VDSi,options); % find the VDS value for these conditions
        
        COND=(EXITFLAG<1);
            if(EXITFLAG>0) 
        % if max(abs(FVAL))<1e-6
                VDSm(1,ip)=VDS(1);
                VDSp(1,ip)=VDS(2);
                VDSi=VDS;
            else
                VDSi(1,1)=VDSi(1,1)+0.2*(rand(1,1)*2-1+1i*(rand(1,1)*2-1));
                VDSi(2,1)=VDSi(2,1)+0.2*(rand(1,1)*2-1+1i*(rand(1,1)*2-1));
                itercnt=itercnt+1;
                if(itercnt>1000)
                    itercnt=1;
                end
            end
        end

        % DC currents of main and peaking amplilfiers
        I0m(1,ip)=DrainCurrent1m(Ipm,IDQm,((1-ConvDoh)*exp(1i*pi/2*w(ip)/wN)+ConvDoh)*VDS(1),No,nchko);    % no phase shift needed
        I0m(1,ip)=(I0m(1,ip)+DrainCurrent1(Ipm,IDQm,((1-ConvDoh)*exp(1i*pi/2*w(ip)/wN)+ConvDoh)*VDS(1),Ne))/2;    % no phase shift needed

        I0p(1,ip)=DrainCurrent1m(Ipp,IDQp,(1-ConvDoh+ConvDoh*exp(1i*pi/2*w(ip)/wN))*VDS(2),No,nchko); % phase shift VDS(2) as if it is cosine excited
        I0p(1,ip)=(I0p(1,ip)+DrainCurrent1(Ipp,IDQp,(1-ConvDoh+ConvDoh*exp(1i*pi/2*w(ip)/wN))*VDS(2),Ne))/2; % phase shift VDS(2) as if it is cosine excited

        % fundamental currents of main and peaking amplifiers
        IDSm(1,ip)=(ConvDoh+(1-ConvDoh)*exp(-1i*pi/2*w(ip)/wN))*DrainCurrent3m(Ipm,IDQm,((1-ConvDoh)*exp(1i*pi/2*w(ip)/wN)+ConvDoh)*VDS(1),No,nchko); % current of main trans.
        IDSm(1,ip)=(IDSm(1,ip)+(ConvDoh+(1-ConvDoh)*exp(-1i*pi/2*w(ip)/wN))*DrainCurrent3(Ipm,IDQm,((1-ConvDoh)*exp(1i*pi/2*w(ip)/wN)+ConvDoh)*VDS(1),Ne,nchke))/2; % current of main trans. 

        IDSp(1,ip)=(ConvDoh*exp(-1i*pi/2*w(ip)/wN)+1-ConvDoh)*DrainCurrent3m(Ipp,IDQp,(1-ConvDoh+ConvDoh*exp(1i*pi/2*w(ip)/wN))*VDS(2),No,nchko); % current of peaking trans.
        IDSp(1,ip)=(IDSp(1,ip)+(ConvDoh*exp(-1i*pi/2*w(ip)/wN)+1-ConvDoh)*DrainCurrent3(Ipp,IDQp,(1-ConvDoh+ConvDoh*exp(1i*pi/2*w(ip)/wN))*VDS(2),Ne,nchke))/2; % current of peaking trans.

        
        Zm(1,ip)=VDD*VDSm(1,ip)/IDSm(1,ip);
        Zp(1,ip)=VDD*VDSp(1,ip)/IDSp(1,ip);
        
    end
    
    VL=B11a.*IDSm+B12a.*IDSp;
    POUT=1/2*real((VL).*conj(VL./ZL));
    EFF(ii,:)=POUT./((I0m+I0p)*VDD);


    PDB(ii,:)=10*log10(POUT)+30;
    toc
    stepNo=stepNo+1;
    figure(4)

    subplot(1,2,1)
    h=plot(fn,PDB,'LineWidth',2);
    %hold on

     grid on
     xlabel('f (GHz)')
     ylabel('P_{out} (dBm)')
    if(wmin<wmax)
        axis([wmin/(2*pi*1e9),wmax/(2*pi*1e9),35,45]);
    else
        axis([wmax/(2*pi*1e9),wmin/(2*pi*1e9),35,45]);
    end
    
    subplot(1,2,2)
    h1=plot(fn,100*EFF,'LineWidth',2);

    grid on
    ylabel('\eta (%)')
    xlabel('f (GHz)')
    if(wmin<wmax)
        axis([wmin/(2*pi*1e9),wmax/(2*pi*1e9),40,80]);
    else
        axis([wmax/(2*pi*1e9),wmin/(2*pi*1e9),40,80]);
    end

    figure(1)
    subplot(1,2,1)
    
    
    
    thick=3;
    
    gamma = z2gamma(Zm(1:nw));
    [line,hsm]=smithchart(gamma);
    hold on
    if(ii==1)
        set(line,'LineWidth',thick,'Color','black','LineStyle','-');
    else
        set(line,'LineWidth',thick,'Color','blue','LineStyle','-');
    end
    
    if(ii>1)
        gamma= z2gamma(Zp(1:nw));
        [line,hsm]=smithchart(gamma);
        set(line,'LineWidth',thick,'Color','red','LineStyle','-');
    end
    
    subplot(1,2,2)
    
    hold on
    
    thick=3;
    
    gamma = z2gamma(Zm(1:nw));
    hold on
    [line,hsm]=smithchart(gamma);
    
    if(ii==1)
        set(line,'LineWidth',thick,'Color','black','LineStyle','-');
    else
        set(line,'LineWidth',thick,'Color','blue','LineStyle','-');
    end
    
    
    
    if(ii>1)
        gamma= z2gamma(Zp(1:nw));
        [line,hsm]=smithchart(gamma);
        set(line,'LineWidth',thick,'Color','red','LineStyle','-');
    end
    

end

function f=sA1(Ipm,Ipp,IDQm,IDQp,VDS,No,nchko,Ne,nchke,AT,wp,VDD,ConvDoh,y)
% 
    IDSm=(ConvDoh+(1-ConvDoh)*exp(-1i*2*y*pi*wp))*DrainCurrent3m(Ipm,IDQm,((1-ConvDoh)*exp(1i*2*y*pi*wp)+ConvDoh)*VDS(1),No,nchko);
    IDSm=(IDSm+(ConvDoh+(1-ConvDoh)*exp(-1i*2*y*pi*wp))*DrainCurrent3(Ipm,IDQm,((1-ConvDoh)*exp(1i*2*y*pi*wp)+ConvDoh)*VDS(1),Ne,nchke))/2;

    IDSp=(ConvDoh*exp(-1i*2*y*pi*wp)+1-ConvDoh)*DrainCurrent3m(Ipp,IDQp,(1-ConvDoh+ConvDoh*exp(1i*2*y*pi*wp))*VDS(2),No,nchko);
    IDSp=(IDSp+(ConvDoh*exp(-1i*2*y*pi*wp)+1-ConvDoh)*DrainCurrent3(Ipp,IDQp,(1-ConvDoh+ConvDoh*exp(1i*2*y*pi*wp))*VDS(2),Ne,nchke))/2;
    
f=VDD*VDS-AT*[IDSm;IDSp]; % vector to be zeroed
end
